@@ -59,12 +59,16 @@
 extern "C" {
   #include <libnetfilter_conntrack/libnetfilter_conntrack.h>
 };
+#define __STDC_FORMAT_MACROS
+#include <inttypes.h>
+#include <locale.h>
 #include <netdb.h>
 #include <ncurses.h>
 #include <unistd.h>
+#include <sys/ioctl.h>
 using namespace std;
 
-#define VERSION "2.2.6"
+#define VERSION "2.2.7"
 /*
  * MAXCONS is set to 16k, the default number of states in iptables. Generally
  * speaking the ncurses pad is this many lines long, but since ncurses
@@ -149,7 +153,7 @@ struct screensize_t {
 // Struct 'o flags
 struct flags_t {
   bool single, totals, lookup, skiplb, staticsize, skipdns, tag_truncate,
-       filter_src, filter_dst, filter_srcpt, filter_dstpt, noscroll, nocolor,
+       filter_src, filter_dst, filter_srcpt, filter_dstpt, filter_inv, noscroll, nocolor,
        counters;
 };
 // Struct 'o counters
@@ -159,6 +163,8 @@ struct counters_t {
 // Various filters to be applied pending the right flags in flags_t
 struct filters_t {
   in6_addr src, dst;
+  bool has_srcnet, has_dstnet;
+  uint8_t srcnet, dstnet;
   uint8_t srcfam, dstfam;
   unsigned long srcpt, dstpt;
 };
@@ -222,24 +228,88 @@ unsigned int digits(unsigned long x)
 }
 
 /*
- * Check to ensure an IP is valid
+ * Check to ensure an IP & netmask are valid
  */
-bool check_ip(const char *arg, in6_addr *addr, uint8_t *family)
+bool check_ip(const char *arg, in6_addr *addr, uint8_t *family, uint8_t *netmask, bool *has_netmask)
 {
+  const char *p_arg;
+  char ip[NAMELEN];
+
+  // Check for netmask prefix
+  p_arg = strrchr(arg, '/');
+  if (p_arg == NULL) {
+    memcpy(ip, arg, (strlen(arg) + 1));
+    *has_netmask = false;
+  } else {
+    size_t ip_len = strlen(arg) - strlen(p_arg);
+    memcpy(ip, arg, ip_len);
+    ip[ip_len] = '\0';
+
+    const char *net_arg = arg + ip_len + 1;
+    if (net_arg[0] == '\0')
+      return false;
+    int net_int = atoi(net_arg);
+    if (net_int < 0 || net_int > 128)
+      return false; // Max IPv6 prefix length
+    *has_netmask = true;
+    *netmask = (uint8_t) net_int;
+  }
+
+  // Get IP
   int ret;
-  ret = inet_pton(AF_INET6, arg, addr);
+  ret = inet_pton(AF_INET6, ip, addr);
   if (ret) {
     *family = AF_INET6;
     return true;
   }
-  ret = inet_pton(AF_INET, arg, addr);
+  ret = inet_pton(AF_INET, ip, addr);
   if (ret) {
+    if (*has_netmask && *netmask > 32)
+      return false; // Max IPv4 prefix length
     *family = AF_INET;
     return true;
   }
   return false;
 }
 
+/*
+ * Compare IPv4/6 addresses with a netmask 
+ * https://gist.github.com/duedal/b83303b4988a4afb2a75
+ */
+bool match_netmask(uint8_t family, const in6_addr &address, const in6_addr &network, uint8_t bits) {
+  if (family == AF_INET) {
+    if (bits == 0) {
+      // C99 6.5.7 (3): u32 << 32 is undefined behaviour
+      return true;
+    }
+
+    struct in_addr addr4, net4;
+    memcpy(&addr4, &address, sizeof(in_addr));
+    memcpy(&net4, &network, sizeof(in_addr));
+    return !((addr4.s_addr ^ net4.s_addr) & htonl(0xFFFFFFFFu << (32 - bits)));
+  } else {
+    const uint32_t *a = address.s6_addr32;
+    const uint32_t *n = network.s6_addr32;
+
+    int bits_whole, bits_incomplete;
+    bits_whole = bits >> 5;         // number of whole u32
+    bits_incomplete = bits & 0x1F;  // number of bits in incomplete u32
+    if (bits_whole) {
+      if (memcmp(a, n, bits_whole << 2)) {
+        return false;
+      }
+    }
+    if (bits_incomplete) {
+      uint32_t mask = htonl((0xFFFFFFFFu) << (32 - bits_incomplete));
+      if ((a[bits_whole] ^ n[bits_whole]) & mask) {
+        return false;
+      }
+    }
+    return true;
+  }
+  return false;
+}
+
 /*
  * The help
  */
@@ -256,14 +326,15 @@ void help()
   cout << "\tToggle color-code by protocol\n\n";
   cout << "  -C, --counters\n";
   cout << "\tToggle display of bytes/packets counters\n\n";
-  cout << "  -d, --dst-filter <IP>\n";
-  cout << "\tOnly show states with a destination of <IP>\n";
-  cout << "\tNote, that this must be an IP, hostname matching is"
-    << " not yet supported.\n\n";
+  cout << "  -d, --dst-filter <IP>[/<NETMASK>]\n";
+  cout << "\tOnly show states with a destination of <IP> and optional <NETMASK>\n";
+  cout << "\tNote: Hostname matching is not yet supported.\n\n";
   cout << "  -D --dstpt-filter <port>\n";
   cout << "\tOnly show states with a destination port of <port>\n\n";
   cout << "  -h, --help\n";
   cout << "\tThis help message\n\n";
+  cout << "  -i, --invert-filters\n";
+  cout << "\tInvert filters to display non-matching results\n\n";
   cout << "  -l, --lookup\n";
   cout << "\tShow hostnames instead of IP addresses. Enabling this will also"
     << " enable\n\t-L to prevent an ever-growing number of DNS requests.\n\n";
@@ -296,10 +367,9 @@ void help()
   cout << "\tNote that bytes/packets are only available when"
     << " supported in the kernel,\n";
   cout << "\tand enabled with -C\n\n";
-  cout << "  -s, --src-filter <IP>\n";
-  cout << "\tOnly show states with a source of <IP>\n";
-  cout << "\tNote, that this must be an IP, hostname matching is"
-    << " not yet supported.\n\n";
+  cout << "  -s, --src-filter <IP>[/<NETMASK>]\n";
+  cout << "\tOnly show states with a source of <IP> and optional <NETMASK>\n";
+  cout << "\tNote: Hostname matching is not yet supported.\n\n";
   cout << "  -S, --srcpt-filter <port>\n";
   cout << "\tOnly show states with a source port of <port>\n\n";
   cout << "  -t, --totals\n";
@@ -586,10 +656,14 @@ screensize_t get_size(const bool &single)
   int maxx = 0, maxy = 0;
   if (!single) {          
     getmaxyx(stdscr, maxy, maxx);
-  } else {                             
-    maxx = 72;
+  } else {
+    // https://stackoverflow.com/questions/1022957/
+    struct winsize w;
+    ioctl(0, TIOCGWINSZ, &w);
+    maxx = w.ws_col;
+
     if (getenv("COLS"))
-      maxx=atoi(getenv("COLS"));
+      maxx = atoi(getenv("COLS"));
   }
 
   screensize_t a;
@@ -682,7 +756,7 @@ void get_input(WINDOW *win, string &input, const string &prompt,
   if (!flags.nocolor)
     wattron(cmd, COLOR_PAIR(4));
   keypad(cmd, true);
-  wprintw(cmd, prompt.c_str());
+  wprintw(cmd, "%s", prompt.c_str());
   wclrtoeol(cmd);
   prefresh(cmd, 0, 0, 0, 0, 0, x);
 
@@ -707,9 +781,13 @@ void get_input(WINDOW *win, string &input, const string &prompt,
         wmove(win, 0, 0);
         return;
         break;
-      // 8 is shift-backspace - just incase
+      // on most platforms KEY_BACKSPACE will catch
+      // all backspaces...
       case KEY_BACKSPACE:
+      // but on some platforms ncurses fails, so ensure we catch both 8 (0x8) and 127 (0x7e)
+      // which are the two valid backspace keycodes
       case 8:
+      case 127:
         if (charcount > 0) {
           input = input.substr(0, input.size()-1);
           wechochar(cmd, '\b');
@@ -747,7 +825,7 @@ void c_warn(WINDOW *win, const string &warning, const flags_t &flags)
   WINDOW *warn = subpad(win, 1, x, 0, 0);
   if (!flags.nocolor)
     wattron(warn, COLOR_PAIR(4));
-  wprintw(warn, warning.c_str());
+  wprintw(warn, "%s", warning.c_str());
   wprintw(warn, " Press any key to continue...");
   wclrtoeol(warn);
   prefresh(warn, 0, 0, 0, 0, 0, x);
@@ -920,7 +998,7 @@ void delete_state(WINDOW *&win, const tentry_t *entry, const flags_t &flags)
   if (ret < 0) {
     string msg = "Failed to delete state: ";
     msg += strerror(errno);
-    c_warn(win, msg.c_str(), flags);
+    c_warn(win, msg, flags);
   }
 
 }
@@ -999,7 +1077,7 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
   minutes = minutes%60;
   seconds = seconds%60;
   // Format it with snprintf and store it in the table
-  snprintf(ttlc,11, "%3i:%02i:%02i", hours, minutes, seconds);
+  snprintf(ttlc, 11, "%3i:%02i:%02i", hours, minutes, seconds);
   entry->ttl = ttlc;
 
   entry->family = nfct_get_attr_u8(ct, ATTR_ORIG_L3PROTO);
@@ -1086,24 +1164,52 @@ int conntrack_hook(enum nf_conntrack_msg_type nf_type, struct nf_conntrack *ct,
     return NFCT_CB_CONTINUE;
   }
 
-  if (flags->filter_src && (memcmp(&(entry->src), &(filters->src), entrysize))) {
-    counts->skipped++;
-    return NFCT_CB_CONTINUE;
+  if (flags->filter_src && !filters->has_srcnet) {
+    if ((flags->filter_inv && !memcmp(&(entry->src), &(filters->src), entrysize)) || 
+        (!flags->filter_inv && memcmp(&(entry->src), &(filters->src), entrysize))) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
   }
 
-  if (flags->filter_srcpt && (entry->srcpt != filters->srcpt)) {
-    counts->skipped++;
-    return NFCT_CB_CONTINUE;
+  if (flags->filter_src && filters->has_srcnet) {
+    if ((flags->filter_inv && match_netmask(entry->family, entry->src, filters->src, filters->srcnet)) || 
+        (!flags->filter_inv && !match_netmask(entry->family, entry->src, filters->src, filters->srcnet))) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
   }
 
-  if (flags->filter_dst && (memcmp(&(entry->dst), &(filters->dst), entrysize))) {
-    counts->skipped++;
-    return NFCT_CB_CONTINUE;
+  if (flags->filter_srcpt) {
+    if ((flags->filter_inv && entry->srcpt == filters->srcpt) || 
+        (!flags->filter_inv && entry->srcpt != filters->srcpt)) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
   }
 
-  if (flags->filter_dstpt && (entry->dstpt != filters->dstpt)) {
-    counts->skipped++;
-    return NFCT_CB_CONTINUE; 
+  if (flags->filter_dst && !filters->has_dstnet) {
+    if ((flags->filter_inv && !memcmp(&(entry->dst), &(filters->dst), entrysize)) || 
+        (!flags->filter_inv && memcmp(&(entry->dst), &(filters->dst), entrysize))) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
+  }
+
+  if (flags->filter_dst && filters->has_dstnet) {
+    if ((flags->filter_inv && match_netmask(entry->family, entry->dst, filters->dst, filters->dstnet)) || 
+        (!flags->filter_inv && !match_netmask(entry->family, entry->dst, filters->dst, filters->dstnet))) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
+  }
+
+  if (flags->filter_dstpt) {
+    if ((flags->filter_inv && entry->dstpt == filters->dstpt) || 
+        (!flags->filter_inv && entry->dstpt != filters->dstpt)) {
+      counts->skipped++;
+      return NFCT_CB_CONTINUE;
+    }
   }
 
   /*
@@ -1323,6 +1429,12 @@ void print_headers(const flags_t &flags, const string &format,
         printf("src: %s", tmp);
       else
         wprintw(mainwin, "src: %s", tmp);
+      if (filters.has_srcnet) {
+        if (flags.single)
+          printf("/%" PRIu8, filters.srcnet);
+        else
+          wprintw(mainwin, "/%" PRIu8, filters.srcnet);
+      }
       printed_a_filter = true;
     }
     if (flags.filter_srcpt) {
@@ -1350,6 +1462,12 @@ void print_headers(const flags_t &flags, const string &format,
         printf("dst: %s", tmp);
       else
         wprintw(mainwin, "dst: %s", tmp);
+      if (filters.has_dstnet) {
+        if (flags.single)
+          printf("/%" PRIu8, filters.dstnet);
+        else
+          wprintw(mainwin, "/%" PRIu8, filters.dstnet);
+      }
       printed_a_filter = true;
     }
     if (flags.filter_dstpt) {
@@ -1365,6 +1483,13 @@ void print_headers(const flags_t &flags, const string &format,
         wprintw(mainwin, "dport: %lu", filters.dstpt);
       printed_a_filter = true;
     }
+    if (flags.filter_inv) {
+      if (flags.single) {
+        printf(" (Inverted)");
+      } else {
+        wprintw(mainwin, " (Inverted)");
+      }
+    }
     if (flags.single)
       printf("\n");
     else
@@ -1586,7 +1711,10 @@ void determine_format(WINDOW *mainwin, max_t &max, screensize_t &ssize,
          */
   if (ssize.x < 85 && flags.counters && !flags.lookup) {
     string prompt = "Window too narrow for counters! Disabling.";
-    c_warn(mainwin, prompt, flags);
+    if (flags.single)
+      cerr << prompt << endl;
+    else
+      c_warn(mainwin, prompt, flags);
     flags.counters = false;
   }
 
@@ -1701,23 +1829,25 @@ void interactive_help(const string &sorting, const flags_t &flags,
    * big we're going to create the pad.
    * 
    * In many cases we'd make the pad very very large and not
-   * worry about it. However, in this case:
-   *   1. We know exactly how big we need it to be, and it's
-   *      not going to change interactively.
-   *   2. We want to draw a "box" around the window and if the
-   *      pad is huge then the box will get drawn around that.
+   * worry about it. However, in this case, we want to draw 
+   * a "box" around the window and if the pad is huge then 
+   * the box will get drawn around that.
    *
-   * So... we have 32 lines of help, plus a top and bottom border,
-   * thus maxrows is 34.
+   * So... we have 41 lines of help, plus a top and bottom border,
+   * thus maxrows is 43. We also need to account for the filter settings
+   * that are only being displayed when enabled.
    *
-   * Our help text is not wider than 80, so we'll se that standard
+   * Our help text is not wider than 80, so we'll set that standard
    * width.
    *
    * If the screen is bigger than this, we deal with it below.
    */
-  unsigned int maxrows = 41;
+  unsigned int maxrows = 43;
   unsigned int maxcols = 80;
 
+  // Acount for dynamic filter settings
+  maxrows += flags.filter_src + flags.filter_srcpt + flags.filter_dst + flags.filter_dstpt;
+
   /*
    * The actual screen size
    */
@@ -1843,21 +1973,28 @@ void interactive_help(const string &sorting, const flags_t &flags,
   waddstr(helpwin,(flags.counters) ? "yes" : "no");
   wattroff(helpwin, A_BOLD);
 
+  mvwaddstr(helpwin, y++, x, "  Invert filters: ");
+  wattron(helpwin, A_BOLD);
+  waddstr(helpwin,(flags.filter_inv) ? "yes" : "no");
+  wattroff(helpwin, A_BOLD);
+
   char tmp[NAMELEN];
   if (flags.filter_src) {
-    inet_ntop(filters.srcfam, &filters.src, tmp,
-              filters.srcfam == AF_INET ? sizeof(in_addr) : sizeof(in6_addr));
+    inet_ntop(filters.srcfam, &filters.src, tmp, NAMELEN-1);
     mvwaddstr(helpwin, y++, x, "  Source filter: ");
     wattron(helpwin, A_BOLD);
     waddstr(helpwin, tmp);
+    if (filters.has_srcnet)
+      wprintw(helpwin, "/%" PRIu8, filters.srcnet);
     wattroff(helpwin, A_BOLD);
   }
   if (flags.filter_dst) {
-    inet_ntop(filters.dstfam, &filters.dst, tmp,
-              filters.dstfam == AF_INET ? sizeof(in_addr) : sizeof(in6_addr));
+    inet_ntop(filters.dstfam, &filters.dst, tmp, NAMELEN-1);
     mvwaddstr(helpwin, y++, x, "  Destination filter: ");
     wattron(helpwin, A_BOLD);
     waddstr(helpwin, tmp);
+    if (filters.has_dstnet)
+      wprintw(helpwin, "/%" PRIu8, filters.dstnet);
     wattroff(helpwin, A_BOLD);
   }
   if (flags.filter_srcpt) {
@@ -1918,6 +2055,11 @@ void interactive_help(const string &sorting, const flags_t &flags,
   wattroff(helpwin, A_BOLD);
   waddstr(helpwin, "\tDisplay this help message");
 
+  wattron(helpwin, A_BOLD);
+  mvwaddstr(helpwin, y++, x, "  i");
+  wattroff(helpwin, A_BOLD);
+  waddstr(helpwin, "\tInvert filters to display non-matching results");
+
   wattron(helpwin, A_BOLD);
   mvwaddstr(helpwin, y++, x, "  l");
   wattroff(helpwin, A_BOLD);
@@ -2122,6 +2264,8 @@ out:
  */
 int main(int argc, char *argv[])
 {
+  // Use the locale specified by the environment
+  setlocale(LC_ALL, "");
 
   // Variables
   string line, src, dst, srcpt, dstpt, proto, code, type, state, ttl, mins,
@@ -2144,7 +2288,7 @@ int main(int argc, char *argv[])
   flags.single = flags.totals = flags.lookup = flags.skiplb = flags.staticsize
       = flags.skipdns = flags.tag_truncate = flags.filter_src
       = flags.filter_dst = flags.filter_srcpt = flags.filter_dstpt
-      = flags.noscroll = flags.nocolor = flags.counters = false;
+      = flags.noscroll = flags.nocolor = flags.counters = flags.filter_inv = false;
   ssize.x = ssize.y = 0;
   counts.tcp = counts.udp = counts.icmp = counts.other = counts.skipped = 0;
   filters.src = filters.dst = in6addr_any;
@@ -2157,6 +2301,7 @@ int main(int argc, char *argv[])
     {"dst-filter", required_argument, 0, 'd'},
     {"dstpt-filter", required_argument, 0, 'D'},
     {"help", no_argument, 0, 'h'},
+    {"invert-filters", no_argument, 0, 'i'},
     {"lookup", no_argument, 0, 'l'},
     {"mark-truncated", no_argument, 0, 'm'},
     {"no-color", no_argument, 0, 'c'},
@@ -2177,9 +2322,8 @@ int main(int argc, char *argv[])
   int option_index = 0;
 
   // Command Line Arguments
-  while ((tmpint = getopt_long(argc, argv, "Cd:D:hlmcoLfpR:r1b:s:S:tv",
+  while ((tmpint = getopt_long(argc, argv, "Cd:D:hilmcoLfpR:r1b:s:S:tv",
                                long_options, &option_index)) != EOF) {
-    printf("loop: %d\n", tmpint);
     switch (tmpint) {
     case 0:
       /* Apparently this test is needed?! Seems lame! */
@@ -2208,7 +2352,7 @@ int main(int argc, char *argv[])
       if (optarg == NULL)
         break;
       // See check_ip() note above
-      if (!check_ip(optarg, &filters.dst, &filters.dstfam)) {
+      if (!check_ip(optarg, &filters.dst, &filters.dstfam, &filters.dstnet, &filters.has_dstnet)) {
         cerr << "Invalid IP address: " << optarg
           << endl;
         exit(1);
@@ -2227,6 +2371,10 @@ int main(int argc, char *argv[])
       flags.filter_dstpt = true;
       filters.dstpt = atoi(optarg);
       break;
+    // --invert-filters
+    case 'i':
+      flags.filter_inv = true;
+      break;
     // --help
     case 'h':
       help();
@@ -2296,7 +2444,7 @@ int main(int argc, char *argv[])
     case 's':
       if (optarg == NULL)
         break;
-      if (!check_ip(optarg, &filters.src, &filters.srcfam)) {
+      if (!check_ip(optarg, &filters.src, &filters.srcfam, &filters.srcnet, &filters.has_srcnet)) {
         cerr << "Invalid IP address: " << optarg << endl;
         exit(1);
       }
@@ -2408,7 +2556,10 @@ int main(int argc, char *argv[])
       prompt = "Counters requested, but not enabled in the";
       prompt += " kernel!";
       flags.counters = 0;
-      c_warn(mainwin, prompt, flags);
+      if (flags.single)
+        cerr << prompt << endl;
+      else
+        c_warn(mainwin, prompt, flags);
     }
 
     // Sort our table
@@ -2483,6 +2634,9 @@ int main(int argc, char *argv[])
       case 'h':
         interactive_help(sorting, flags, filters);
         break;
+      case 'i':
+        flags.filter_inv = !flags.filter_inv;
+        break;
       case 'l':
         flags.lookup = !flags.lookup;
         /*
@@ -2545,7 +2699,7 @@ int main(int argc, char *argv[])
           flags.filter_dst = false;
           filters.dst = in6addr_any;
         } else {
-          if (!check_ip(tmpstring.c_str(), &filters.dst, &filters.dstfam)) {
+          if (!check_ip(tmpstring.c_str(), &filters.dst, &filters.dstfam, &filters.dstnet, &filters.has_dstnet)) {
             prompt = "Invalid IP,";
             prompt += " ignoring!";
             c_warn(mainwin, prompt, flags);
@@ -2590,7 +2744,7 @@ int main(int argc, char *argv[])
           flags.filter_src = false;
           filters.src = in6addr_any;
         } else {
-          if (!check_ip(tmpstring.c_str(), &filters.src, &filters.srcfam)) {
+          if (!check_ip(tmpstring.c_str(), &filters.src, &filters.srcfam, &filters.srcnet, &filters.has_srcnet)) {
             prompt = "Invalid IP,";
             prompt += " ignoring!";
             c_warn(mainwin, prompt, flags);
